# -*- coding: utf-8 -*-
from collections import OrderedDict
import logging
import json
import requests
import traceback

from odoo import models, fields, _
from odoo.addons.saas_worker.const import SCHEME, TIMEOUT
from odoo.api import depends
from odoo.exceptions import AccessDenied
from odoo.tools.misc import split_every
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.metabase import get_port
from odoo.addons.saas_worker.util import get_server_names_by_ip, humanize_version, normalize_ip_address

_logger = logging.getLogger(__name__)


# Loading a new database context can take up to 500 ms if the registry cache is cold
# This value of 200 jobs mean that the execution of a cron should take at most around 100s,
# which is way less than the cron execution time limit, so it should avoid a timeout happening during the execution of a job.
MAX_JOBS_BY_BATCH = 200

# Limit how many databases are treated per call to /saas_worker/multi route to avoid timeouts
MAX_DBS_PER_REQUEST = 20

URL_BY_JOB_TYPE = {
    'invite_users': '{scheme}://{host}/saas_worker/multi/create_invitation/json',
    'remove_users': '{scheme}://{host}/saas_worker/multi/remove_users/json',
}


class OpenerpEnterpriseMultidbJob(models.Model):
    _name = 'openerp.enterprise.multidb.job'
    _description = 'Action on multiple databases'

    user_id = fields.Many2one('res.users', string='Initiator', required=True, ondelete='cascade')
    impostor_id = fields.Many2one('res.users', string='Impostor',
                                  help="If the initiator was impersonated, this field contains the user "
                                       "who is behind this impersonation")
    type = fields.Selection([
        ('invite_users', 'Invitation of users'),
        ('remove_users', 'Removal of users'),
    ], string='Type', required=True)
    arguments = fields.Char(string='Arguments')
    state = fields.Selection(string='Status', compute='_compute_state', index=True, precompute=True, required=True, store=True,
                             selection=[
                                 ('todo',  'In progress'),
                                 ('error', 'Error'),
                                 ('done',  'Done'),
                             ])
    result = fields.Text(string='Result', readonly=True)
    line_ids = fields.One2many('openerp.enterprise.multidb.job.line', 'job_id', string='Job Lines')

    @depends('line_ids.state', 'result')
    def _compute_state(self):
        for job in self:
            if job.state == 'error':
                # finalize may have failed
                continue
            elif any(line.state == 'error' for line in job.line_ids):
                job.state = 'error'
            else:
                job.state = 'done' if job.result else 'todo'

    def _compute_display_name(self):
        type2str = dict(self._fields['type']._description_selection(self.env))
        for record in self:
            record.display_name = type2str[record.type]

    def _cron_process_jobs(self):
        server_names_by_ip = get_server_names_by_ip()

        nb_jobs_remaining = MAX_JOBS_BY_BATCH

        jobs = self.search([('state', '=', 'todo')])
        lines_by_ip_by_job = OrderedDict()
        for job in jobs:
            lines_by_ip = lines_by_ip_by_job.setdefault(job, {})
            for line in job.line_ids.filtered(lambda l: l.state == 'todo'):
                server_ip = normalize_ip_address(line.database_id.last_ip)
                if server_ip in server_names_by_ip:
                    version = humanize_version(line.database_id.version)
                    serverport = (server_ip, get_port(version))
                    try:
                        lines_by_ip[serverport] |= line
                    except KeyError:
                        lines_by_ip[serverport] = line
                else:
                    line.write({
                        'state': 'db_not_found',
                        'result': f'unknown hostname for IP {line.database_id.last_ip}',
                    })

        for job, lines_by_ip in lines_by_ip_by_job.items():
            if nb_jobs_remaining <= 0:
                break

            jobtype = job.type
            try:
                for serverport, lines in lines_by_ip.items():
                    if nb_jobs_remaining <= 0:
                        break
                    if len(lines) > nb_jobs_remaining:
                        lines = lines[:nb_jobs_remaining]
                    nb_jobs_remaining -= len(lines)
                    dbnames = lines.mapped('database_id').mapped('db_name')
                    try:
                        # _run methods are responsible for describing errors in lines.result
                        # and updating the lines' state
                        lines._run(*serverport)
                    except Exception:
                        self.env.cr.rollback()
                        _logger.exception('An error occurred while processing multidb job %s %r on host %s:%s (dbs %r)',
                                          jobtype, job, server_names_by_ip[serverport[0]], serverport[1], dbnames)
                        lines.write({
                            'state': 'error',
                            'result': traceback.format_exc(),
                        })
                    finally:
                        self.env.cr.commit()

                if job.line_ids.went_well():
                    # only finalize if all the runs went well
                    try:
                        # _finalize method are responsible for describing result or error in job.result
                        # and updating the job's state
                        job.line_ids._finalize()
                    finally:
                        if not job.result:
                            job.write({
                                'state': 'error',
                                'result': 'Field `result` should have been filled up',
                            })
            except Exception:
                self.env.cr.rollback()
                _logger.exception('An error occurred while processing multidb job %s %r', jobtype, job)
                job.write({
                    'state': 'error',
                    'result': traceback.format_exc(),
                })
            finally:
                if job.state == 'error':
                    # Send an error notification to the initiator
                    error_template = self.env.ref('openerp_enterprise.multidb_job_error_email')
                    assert error_template._name == 'mail.template'
                    error_template.send_mail(job.id, email_layout_xmlid='mail.mail_notification_light',
                                             raise_exception=True)

                self.env.cr.commit()

        if self.search([('state', '=', 'todo')]):
            self.trigger()

    def action_retry(self):
        if not self.env.user.has_group('openerp_enterprise.group_support_technical'):
            raise AccessDenied('Restricted to technical support staff')

        def erroneous(s):
            return s.state == 'error'

        erroneous_self = self.filtered(erroneous).sudo()
        erroneous_self.line_ids.filtered(erroneous).write({'state': 'todo'})
        erroneous_self.write({'state': 'todo'})

        if erroneous_self:
            self.trigger()

    def trigger(self):
        self.env.ref('openerp_enterprise.cron_multidb_jobs')._trigger()


class OpenerpEnterpriseMultidbJobLine(models.Model):
    _name = 'openerp.enterprise.multidb.job.line'
    _description = 'Line of an action on multiple databases'

    job_id = fields.Many2one('openerp.enterprise.multidb.job', string='Job', index=True, required=True, ondelete='cascade')
    database_id = fields.Many2one('openerp.enterprise.database', string='Database', required=True)
    state = fields.Selection([
        ('todo',  'In progress'),
        ('done',  'Done'),
        ('db_not_found',  'Database not found'),
        ('error', 'Error'),
    ], string='Status', required=True, default='todo')
    result = fields.Text(string='Result', readonly=True)

    def went_well(self):
        return all(x in {'db_not_found', 'done'} for x in self.mapped('state'))

    def _prepare_databases_parameter(self):
        OAuth2TokenSudo = self.env['auth.oauth2.token'].sudo()

        job_id = self.mapped('job_id')
        job_id.ensure_one()
        user_id = job_id.user_id

        databases = []
        for line in self:
            db_id = line.database_id
            db_user_id = db_id.database_user_ids.filtered(lambda u: u.user_id.id == user_id.id)
            session_info = {
                'impostor': job_id.impostor_id.id,
                'login': job_id.user_id.login,
                'uid': job_id.user_id.id,
            }
            databases.append({
                'db_uuid': db_id.db_uuid,
                # tokens are short-lived for security reasons
                'oauth_token': OAuth2TokenSudo._get_trial_token(user_id.id, db_id.db_uuid, session_info=session_info),
                'user_login': db_user_id.login,
            })
        return databases

    def _run(self, server_ip, port):
        job_id = self.mapped('job_id')
        job_id.ensure_one()
        method = getattr(self, '_run_' + job_id.type, False)
        method(server_ip, port)

    def _run_multidb_request(self, server_ip, port, job_type, prepare_data):
        job_id = self.mapped('job_id')
        job_id.ensure_one()

        for subself in split_every(MAX_DBS_PER_REQUEST, self.ids, self.browse):
            subself._run_multidb_request_batch(server_ip, port, job_type, prepare_data)

    def _run_multidb_request_batch(self, server_ip, port, job_type, prepare_data):
        job_id = self.mapped('job_id')
        job_id.ensure_one()
        assert len(self) <= MAX_DBS_PER_REQUEST, f"len({self}) > {MAX_DBS_PER_REQUEST}"

        # we need to commit after generating the tokens, so that they are found by the other worker
        # when the trial server will get back to us
        databases = self._prepare_databases_parameter()
        self.env.cr.commit()

        server_names_by_ip = get_server_names_by_ip()
        server_name = server_names_by_ip[server_ip]
        url = URL_BY_JOB_TYPE[job_type].format(scheme=SCHEME, host=server_name)

        arguments = json.loads(job_id.arguments)
        headers = {'X-Odoo-Port': str(port)}
        data = prepare_data(databases, arguments)
        resp = requests.post(url, headers=headers, json=data, auth=SaasAuth(), timeout=len(databases) * TIMEOUT)
        resp.raise_for_status()
        resp_json = resp.json()
        resp_error = resp_json.get('error', False)
        if resp_error:
            self.write({
                'state': 'error',
                'result': json.dumps(resp_error),
            })
            return

        result = resp_json.get('result', {})
        if not result:
            self.write({
                'state': 'error',
                'result': json.dumps(resp_json),
            })
            return

        for line in self:
            if line.database_id.db_uuid not in result:
                line.write({
                    'state': 'db_not_found',
                    'result': json.dumps({'result': False, 'error_code': 'db_not_found'}),
                })
                continue

            db_result = result[line.database_id.db_uuid]
            if not db_result or db_result.get('result', False) is False:
                state = 'error'
                if db_result.get('error_code') == 'db_not_found':
                    state = 'db_not_found'
                line.write({
                    'state': state,
                    'result': json.dumps(db_result),
                })
                continue

            line.write({
                'state': 'done',
                'result': json.dumps(db_result.get('result')),
            })

    def _run_invite_users(self, server_ip, port):
        def prepare_data(databases, arguments):
            invited_users = self.env['res.users'].search([('login', 'in', arguments['invited_users'])], order='id')
            return {'params': {
                'databases': databases,
                'invitees': [{'name': u.name, 'mail': u.email, 'oauth_uid': u.id} for u in invited_users],
            }}

        self._run_multidb_request(server_ip, port, 'invite_users', prepare_data)

    def _run_remove_users(self, server_ip, port):
        def prepare_data(databases, arguments):
            return {'params': {
                'databases': databases,
                'user_oauth_ids': arguments['removed_uids'],
            }}

        self._run_multidb_request(server_ip, port, 'remove_users', prepare_data)

    def _finalize(self):
        job_id = self.mapped('job_id')
        job_id.ensure_one()
        method = getattr(self, '_finalize_' + job_id.type, False)
        method()

    def _finalize_invite_users(self):
        job_id = self.mapped('job_id')
        job_id.ensure_one()

        arguments = json.loads(job_id.arguments)

        db_names = [line.database_id.db_name for line in self if line.state == 'done']
        db_not_found_names = [line.database_id.db_name for line in self if line.state == 'db_not_found']
        invited_users = self.env['res.users'].sudo().search([('login', 'in', arguments['invited_users'])])

        if db_names:
            # Send the invitation notification to the invitees
            invitation_template = self.env.ref('openerp_enterprise.multidb_invitation_email')
            assert invitation_template._name == 'mail.template'

            # make sure that the recipient is not polluted, for security reasons
            template_values = {
                'email_to': "{{ctx['user'].email_formatted}}",
                'email_cc': False,
                'auto_delete': True,
                'partner_to': False,
                'scheduled_date': False,
            }
            if any(invitation_template[field] != value for (field, value) in template_values.items()):
                invitation_template.write(template_values)

            for user in invited_users:
                signup = user.login_date is False
                if signup:
                    user.partner_id.signup_prepare()
            # Emails will be sent immediately, so we need to ensure the signup tokens won't be rollbacked
            self.env.cr.commit()

            with self.env.cr.savepoint() as sp:  # Ensure no email containing a signup token is left in the database
                failed_invitations = []
                for user in invited_users:
                    signup = user.login_date is False
                    try:
                        # Depending on signup, the mail will contain an invitation link or not
                        invitation_template.with_context({
                            'db_names': db_names,
                            'user': user,
                            'signup': signup,
                        }).send_mail(job_id.id, email_layout_xmlid='mail.mail_notification_light',
                                     raise_exception=True, force_send=True)
                    except Exception as e:
                        failed_invitations.append((user, e))
                        sp.rollback()

                if failed_invitations:
                    msg = "Failed to send invitations to %s" % (', '.join(u.email for u, e in failed_invitations))
                    raise Exception(msg) from failed_invitations[0][1]

        # Send a confirmation to the inviter
        confirmation_template = self.env.ref('openerp_enterprise.multidb_invitation_confirmation_email')
        assert confirmation_template._name == 'mail.template'
        confirmation_template.with_context({
            'db_names': db_names,
            'db_not_found_names': db_not_found_names,
            'invited_users': invited_users.mapped('login'),
        }).send_mail(job_id.id, email_layout_xmlid='mail.mail_notification_light', raise_exception=True)

        job_id.write({'result': 'true'})

    def _finalize_remove_users(self):
        job_id = self.mapped('job_id')
        job_id.ensure_one()

        arguments = json.loads(job_id.arguments)

        db_names = [line.database_id.db_name for line in self if line.state == 'done']
        db_not_found_names = [line.database_id.db_name for line in self if line.state == 'db_not_found']

        removed_users = self.env['res.users'].sudo().browse([int(i) for i in arguments['removed_uids']])

        # Send a confirmation to the remover
        confirmation_template = self.env.ref('openerp_enterprise.multidb_user_removal_confirmation_email')
        assert confirmation_template._name == 'mail.template'
        confirmation_template.with_context({
            'db_names': db_names,
            'db_not_found_names': db_not_found_names,
            'removed_users': removed_users.mapped('login'),
        }).send_mail(job_id.id, email_layout_xmlid='mail.mail_notification_light', raise_exception=True)

        job_id.write({'result': 'true'})
